﻿// Copyright (c) AlphaSierraPapa for the SharpDevelop Team (for details please see \doc\copyright.txt)
// This code is distributed under the GNU LGPL (for details please see \doc\license.txt)

using System;
using System.Collections.Generic;
using System.Threading;
using System.Windows.Threading;

using ICSharpCode.AvalonEdit.AddIn.Options;
using ICSharpCode.Core;
using ICSharpCode.SharpDevelop;
using ICSharpCode.SharpDevelop.Dom;
using ICSharpCode.SharpDevelop.Editor;
using ICSharpCode.SharpDevelop.Gui;
using ICSharpCode.SharpDevelop.Refactoring;

namespace ICSharpCode.AvalonEdit.AddIn
{
	/// <summary>
	/// In the code editor, highlights all references to the expression under the caret (for better code readability).
	/// </summary>
	public class CaretReferencesRenderer
	{
		/// <summary>
		/// Delays the highlighting after the caret position changes, so that Find references does not get called too often.
		/// </summary>
		DispatcherTimer delayTimer;
		const int delayMilliseconds = 800;
		/// <summary>
		/// Delays the Resolve check so that it does not get called too often when user holds an arrow.
		/// </summary>
		DispatcherTimer delayMoveTimer;
		const int delayMoveMilliseconds = 100;
		
		CodeEditorView editorView;
		ITextEditor Editor { get { return editorView.Adapter; } }
		
		ExpressionHighlightRenderer highlightRenderer;
		ResolveResult lastResolveResult;
		
		public bool IsEnabled
		{
			get {
				string fileName = this.Editor.FileName;
				return CodeEditorOptions.Instance.HighlightSymbol && (fileName.EndsWith(".cs") || fileName.EndsWith(".vb"));
			}
		}
		
		/// <summary>
		/// In the code editor, highlights all references to the expression under the caret (for better code readability).
		/// </summary>
		public CaretReferencesRenderer(CodeEditorView editorView)
		{
			this.editorView = editorView;
			this.highlightRenderer = new ExpressionHighlightRenderer(this.editorView.TextArea.TextView);
			this.delayTimer = new DispatcherTimer() { Interval = TimeSpan.FromMilliseconds(delayMilliseconds) };
			this.delayTimer.Stop();
			this.delayTimer.Tick += TimerTick;
			this.delayMoveTimer = new DispatcherTimer() { Interval = TimeSpan.FromMilliseconds(delayMoveMilliseconds) };
			this.delayMoveTimer.Stop();
			this.delayMoveTimer.Tick += TimerMoveTick;
			this.editorView.TextArea.Caret.PositionChanged += CaretPositionChanged;
		}
		
		public void ClearHighlight()
		{
			this.highlightRenderer.ClearHighlight();
		}

		void TimerTick(object sender, EventArgs e)
		{
			this.delayTimer.Stop();
			
			if (!IsEnabled)
				return;
			var referencesToBeHighlighted = GetReferencesInCurrentFile(this.lastResolveResult);
			this.highlightRenderer.SetHighlight(referencesToBeHighlighted);
		}
		
		void TimerMoveTick(object sender, EventArgs e)
		{
			this.delayMoveTimer.Stop();
			this.delayTimer.Stop();
			
			if (!IsEnabled)
				return;
			
			var resolveResult = GetExpressionUnderCaret();
			if (resolveResult == null) {
				this.lastResolveResult = resolveResult;
				this.highlightRenderer.ClearHighlight();
				return;
			}
			// caret is over symbol and that symbol is different from the last time
			if (!SameResolveResult(resolveResult, lastResolveResult))
			{
				this.lastResolveResult = resolveResult;
				this.highlightRenderer.ClearHighlight();
				this.delayTimer.Start();
			}
		}
		
		/// <summary>
		/// In the current document, highlights all references to the expression
		/// which is currently under the caret (local variable, class, property).
		/// This gets called on every caret position change, so quite often.
		/// </summary>
		void CaretPositionChanged(object sender, EventArgs e)
		{
			this.delayMoveTimer.Stop();
			this.delayMoveTimer.Start();
		}
		
		/// <summary>
		/// Resolves the current expression under caret.
		/// This gets called on every caret position change, so quite often.
		/// </summary>
		ResolveResult GetExpressionUnderCaret()
		{
			if (string.IsNullOrEmpty(Editor.FileName) || ParserService.LoadSolutionProjectsThreadRunning)
				return null;
			int line = Editor.Caret.Position.Line;
			int column = Editor.Caret.Position.Column;
			return ParserService.Resolve(line, column, Editor.Document, Editor.FileName);
		}

		/// <summary>
		/// Finds references to resolved expression in the current file.
		/// </summary>
		List<Reference> GetReferencesInCurrentFile(ResolveResult resolveResult)
		{
			var cancellationTokenSource = new CancellationTokenSource();
			using (new Timer(
				delegate {
					LoggingService.Debug("Aborting GetReferencesInCurrentFile due to timeout");
					cancellationTokenSource.Cancel();
				}, null, 200, Timeout.Infinite)) 
			{
				var progressMonitor = new DummyProgressMonitor();
				progressMonitor.CancellationToken = cancellationTokenSource.Token;
				var references = RefactoringService.FindReferencesLocal(resolveResult, Editor.FileName, progressMonitor);
				if (references == null || references.Count == 0)
					return null;
				return references;
			}
		}
		
		/// <summary>
		/// Returns true if the 2 ResolveResults refer to the same symbol.
		/// </summary>
		bool SameResolveResult(ResolveResult resolveResult, ResolveResult resolveResult2)
		{
			/*if (resolveResult == null && resolveResult2 == null)
				return true;
			if (resolveResult == null && resolveResult2 != null)
				return false;
			if (resolveResult != null && resolveResult2 == null)
				return false;*/
			// TODO determine if 2 ResolveResults refer to the same symbol
			return false;
		}
	}
}
